/**************************************************************************************

        MUSTUXLIB - THE COMMON LIBRARY FOR ALL MUSTUX APPLICATIONS
        AUTHOR : See AUTHORS file for details

        This software is distributed under the terms of the GNU General Public License
        as specified in the COPYING file.

***************************************************************************************/

// #include <string.h>
#include <stdio.h>
#include <unistd.h>
//#include <qstring.h>
//#include <qfile.h>
//#include <qtextstream.h>

#include <qstring.h>
#include "MustuxAudioBus.hh"
#include "MustuxDebugger.hh"
#include "MustuxAudioDeviceMapper.hh"

#ifdef THREADED_TRANSFER

TransferThread::TransferThread  (Bus* pBus)
        {
        bus=pBus;
        transferThreadRunning=false;
        transferring=false;
        }

void TransferThread::run()
        {
        transferThreadRunning=true;
        while (transferThreadRunning)
                {
                while ((transferThreadRunning) && (!transferring))
                        usleep(500);
                if (transferThreadRunning)
                        {
                        QMutex mutex;
                        mutex.lock();
                        int r;
                        if (bus->bytesToTransfer==-1)
                                {
                                r = snd_pcm_writei(bus->pcmHandle, bus->safeTransferBuffer, bus->periodSize);
                                }
                        else
                                {
                                if (bus->bytesToTransfer > bus->periodSize)
                                        {
                                        PMESG3("Trying to transfer more than bus buffer size!");
                                        bus->bytesToTransfer = bus->periodSize;
                                        }
                                bus->partialPeriodSize = bus->bytesToTransfer / bus->blockSize;
                                r = snd_pcm_writei(bus->pcmHandle, bus->safeTransferBuffer, bus->partialPeriodSize);
                                }
                        if (r < 0) // buffer underrun
                                {
                                PMESG3("Underrun on bus %s",(const char*) bus->descriptor.latin1());
                                snd_pcm_prepare(bus->pcmHandle);
                                bus->transferError = MustuxAudioDeviceMapper::ERROR_UNDERRUN;
                                }
                        else
                                bus->transferError=0;
                        /* what the heck is this ?
                        if (r == -EAGAIN)
                                return 1;
                        */
                        bus->bytesTransferred = r * bus->blockSize;
                        transferring = false;
                        mutex.unlock();
                        }
                }
        }

#endif


Bus::Bus(int pParentCardId, int pId, int pStream, QString desc)
        {
        parentCardId= pParentCardId;
        id = pId;
        stream=pStream;
        descriptor=desc;
        for (int m=0; m<MAX_SCB_MODES; m++)
                isValidMode[m]=false;
        status=CLOSED;
        stereoBus=false;
        transferBuffer=(char*) 0;
        transferSize=-1;
        QString spId; spId.setNum(pId+1);
        QString s;
        if (pId==0)
                s="st";
        else if (pId==1)
                s="nd";
        else if (pId==2)
                s="rd";
        else s="th";
        QString br = spId+s+" "+(stream==MustuxAudioDeviceMapper::CAPTURE?"IN":"OUT");
        shortName = MustuxAudioDeviceMapper::get_card_name(parentCardId)+" - "+br;
        fullName = MustuxAudioDeviceMapper::get_card_name(parentCardId)+" - "+br+" ("+descriptor+")";

#ifdef THREADED_TRANSFER
        safeTransferBuffer=(char*) 0;
        transferThread = new TransferThread(this);
#endif

        transferError=0;
        PMESG2("%s Bus %d [ card %d ] : \"%s\" created",stream==MustuxAudioDeviceMapper::CAPTURE?"MustuxAudioDeviceMapper::CAPTURE":"PLABACK", pId,parentCardId,(const char*)fullName.latin1());
        }


Bus::~Bus()
        {
        }


void Bus::probe_valid_modes()
        {
        PENTER3;
        stereoBus=false;
        PMESG2("      Probing valid modes for bus  %d:%d ...",parentCardId,id);
        for (int j=0; j<MAX_SCB_MODES; j++)
                {
                PMESG_START("         Trying %d Hz %d channels %d bits ...",scbModeRate[j],scbModeChannels[j],scbModeBitDepht[j]);
                int r = supports(scbModeRate[j],scbModeBitDepht[j],scbModeChannels[j],stream);
                if (r<0)
                        {
                        const char* em = (const char*) MustuxAudioDeviceMapper::get_error_mesg(r);
                        PMESG_END(" FAILED : %s",em);
                        isValidMode[j]=false;
                        }
                else
                        {
                        PMESG_END(" OK");
                        isValidMode[j]=true;
                        if (scbModeChannels[j]==2)
                                stereoBus=true;
                        }
                }
        PEXIT3;
        }



// THIS IS JUST NOT WORKING ! FIX ME...
int Bus::supports(int pRate, int pBitDepht, int pChannels, int pStream)
        {
        PENTER4;
        format = SND_PCM_FORMAT_S16_LE;
        snd_pcm_hw_params_t* tempHwParams;

        if (pBitDepht==8)
                {
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_8_BITS_BANNED_IN_MUSTUX_APPLICATIONS;
                }
        else if (pBitDepht==16)
                {
                format = SND_PCM_FORMAT_S16_LE;
                }
        else if (pBitDepht==24)
                {
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_BIT_DEPTH_SUPPORT_NOT_IMPLEMENTED_YET;
                }
        else if (pBitDepht==32)
                {
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_BIT_DEPTH_SUPPORT_NOT_IMPLEMENTED_YET;
                }
        else
                {
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_INVALID_BIT_DEPTH;
                }

        if (pStream == MustuxAudioDeviceMapper::CAPTURE)
                {
                if (snd_pcm_open(&pcmHandle, (const char*) descriptor.latin1(),SND_PCM_STREAM_CAPTURE, 0) < 0)
                        {
                        PERROR("Cannot open %s for probing MustuxAudioDeviceMapper::CAPTURE capabilities",(const char*) descriptor.latin1());
                        PEXIT4;
                        return MustuxAudioDeviceMapper::ERROR_CANNOT_OPEN_BUS;
                        }
                }
        else
                {
                if (snd_pcm_open(&pcmHandle, (const char*) descriptor.latin1(),SND_PCM_STREAM_PLAYBACK, 0) < 0)
                        {
                        PERROR("Cannot open %s for probing PLAYBACK capabilities",(const char*) descriptor.latin1());
                        PEXIT4;
                        return MustuxAudioDeviceMapper::ERROR_CANNOT_OPEN_BUS;
                        }
                }

        snd_pcm_hw_params_alloca(&tempHwParams);
        if (snd_pcm_hw_params_any(pcmHandle, tempHwParams) < 0)
                {
                snd_pcm_close(pcmHandle);
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_CANNOT_SET_PARAMS_ANY;
                }

        if (snd_pcm_hw_params_test_rate     (pcmHandle, tempHwParams, pRate, 0)!=0)
                {
                snd_pcm_close(pcmHandle);
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_BUS_DOES_NOT_SUPPORT_THIS_SAMPLE_RATE;
                }

        if (snd_pcm_hw_params_test_format   (pcmHandle, tempHwParams, format)!=0)
                {
                snd_pcm_close(pcmHandle);
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_BUS_DOES_NOT_SUPPORT_THIS_BIT_DEPTH;
                }

        if (snd_pcm_hw_params_test_channels (pcmHandle, tempHwParams, pChannels)!=0)
                {
                snd_pcm_close(pcmHandle);
                PEXIT4;
                return MustuxAudioDeviceMapper::ERROR_BUS_DOES_NOT_SUPPORT_THIS_CHANNELS;
                }

        snd_pcm_close(pcmHandle);
        PEXIT4;
        return 1;
        }




bool Bus::valid_mode(int pRate, int pBitDepht, int pChannels)
        {
        PENTER3;
        for (int m=0; m<MAX_SCB_MODES; m++)
                {
                int ra = scbModeRate[m];
                int bd = scbModeBitDepht[m];
                int ch = scbModeChannels[m];
                if ( (ra==pRate) &&
                (bd==pBitDepht) &&
                (ch==pChannels)
                )
                        {
                        PEXIT3;
                        return isValidMode[m];
                        }
                }
        PEXIT3;
        return false;
        }



int Bus::open(int pRate, int pBitDepht, int whichChannels,bool block)
        {
        PENTER2;
        int channels; // number of channels effectively open ( not the access code MONO,STEREO,ONLY_LEFT,ONLY_RIGHT, AND SO ON...)
        int r;
        format = SND_PCM_FORMAT_S16_LE;
        QString busFullReference = shortName+" ("+descriptor+")";
        const char* cBusFullReference = (const char*) busFullReference.latin1();
        PMESG2("Opening %s bus : %s ...",stream==MustuxAudioDeviceMapper::PLAYBACK?"PLAYBACK":"MustuxAudioDeviceMapper::CAPTURE",cBusFullReference);
        if (stream==MustuxAudioDeviceMapper::CAPTURE)
                {
                if (allocated_for_capture())
                        {
                        PMESG2("Bus seems to be allocated for capture. No need to re-open it.");
                        PEXIT2;
                        return 1;
                        }
                if (stereoBus)
                        {
                        if ((whichChannels==MustuxAudioDeviceMapper::ONLY_LEFT_CHANNEL) && (status==CAPTURING_FROM_LEFT_CHANNEL))
                                {
                                PERROR("One or more channels of bus %s is already in use.",cBusFullReference);
                                PEXIT2;
                                return MustuxAudioDeviceMapper::ERROR_BUSY_BUS;
                                }
                        if ((whichChannels==MustuxAudioDeviceMapper::ONLY_RIGHT_CHANNEL) && (status==CAPTURING_FROM_RIGHT_CHANNEL))
                                {
                                PERROR("One or more channels of bus %s is already in use.", cBusFullReference);
                                PEXIT2;
                                return MustuxAudioDeviceMapper::ERROR_BUSY_BUS;
                                }
                        if ((whichChannels==MustuxAudioDeviceMapper::STEREO) && (status==CAPTURING_FROM_LEFT_CHANNEL || status==CAPTURING_FROM_RIGHT_CHANNEL))
                                {
                                PERROR("One or more channels of bus %s is already in use.",cBusFullReference);
                                PEXIT2;
                                return MustuxAudioDeviceMapper::ERROR_BUSY_BUS;
                                }
                        channels=2;
                        }
                else
                        {
                        if (whichChannels==MustuxAudioDeviceMapper::STEREO)
                                {
                                PERROR("Bus %s does not support STEREO stuff",cBusFullReference);
                                PEXIT2;
                                return MustuxAudioDeviceMapper::ERROR_MONO_BUS;
                                }
                        channels=1;
                        }
                }
        else
                {
                if (allocated_for_playback())
                        {
                        if ((whichChannels!=MustuxAudioDeviceMapper::MONO) && (status!=PLAYING_TO_MONO))
                                {
                                PMESG2("Bus seems to be allocated for playback. No need to re-open it.");
                                PEXIT2;
                                return 0;
                                }
                        PERROR("Bus %s is busy. ",cBusFullReference);
                        PEXIT2;
                        return -1;
                        }
                else
                        {
                        if (whichChannels==MustuxAudioDeviceMapper::MONO)
                                channels=1;
                        else // STEREO, ONLY_LEFT or ONLY_RIGHT
                                channels=2;
                        }
                }

        // not open yet ? so let's try and open it ! :-)
        if (!valid_mode(pRate, pBitDepht, channels))
                {
                PERROR("Invalid mode for bus %s : %d Hz, %d bits, %s",cBusFullReference,pRate,pBitDepht, (channels==1?"MONO":"STEREO"));
                PEXIT2;
                return MustuxAudioDeviceMapper::ERROR_INVALID_BUS;
                }
        const char* cdescriptor = (const char*) descriptor.latin1();
        if (stream==MustuxAudioDeviceMapper::PLAYBACK)
                {
                r = snd_pcm_open(&pcmHandle, cdescriptor, SND_PCM_STREAM_PLAYBACK, block ? 0 : SND_PCM_NONBLOCK);
                if ( r < 0)
                        {
                        PERROR("Cannot open bus %s for PLAYBACK . ALSA ERROR : %s",cBusFullReference,snd_strerror(r));
                        PEXIT2;
                        return MustuxAudioDeviceMapper::ERROR_CANNOT_OPEN_PLAYBACK_BUS;
                        }
                }
        else if (stream==MustuxAudioDeviceMapper::CAPTURE)
                {
                r = snd_pcm_open(&pcmHandle, cdescriptor, SND_PCM_STREAM_CAPTURE, 0);
                if (r < 0)
                        {
                        PERROR("Cannot open %s for MustuxAudioDeviceMapper::CAPTURE . ALSA ERROR : %s",cBusFullReference,snd_strerror(r));
                        PEXIT2;
                        return MustuxAudioDeviceMapper::ERROR_CANNOT_OPEN_CAPTURE_BUS;
                        }
                }


        // some parameters declaration/initialization
        int dir;
        int triedPeriodSize;
        int numPeriods = 2; // IMPROVE ME (?)
        blockSize = channels * (pBitDepht/8);
        int chunckSize = pRate*blockSize;
        int fragmentSize = chunckSize / MustuxAudioDeviceMapper::get_fps();
        snd_pcm_uframes_t minPeriodSize, maxPeriodSize;


        // starting the ALSA crap settings
        snd_pcm_hw_params_alloca(&hwparams);
        r = snd_pcm_hw_params_any(pcmHandle, hwparams);
        if (r < 0)
                {
                PERROR("Cannot configure bus %s. ALSA ERROR : %s ",cBusFullReference,snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        // set the interleaved read/write format
        r = snd_pcm_hw_params_set_access(pcmHandle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED);
        if (r < 0)
                {
                PERROR("Invalid access format to bus %s: %s",cBusFullReference,snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        // set the sample format
        // SHOULD CHECK THE BITDEPTH HERE....
        r = snd_pcm_hw_params_set_format(pcmHandle, hwparams, format);
        if (r < 0)
                {
                PERROR("Invalid sample format to bus %s : %s",cBusFullReference,snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        // set the count of channels
        r = snd_pcm_hw_params_set_channels(pcmHandle, hwparams, channels);
        if (r < 0)
                {
                PERROR("Invalid number of channels (%d) to bus %s : %s",cBusFullReference,  channels, snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        // set the stream rate
        r = snd_pcm_hw_params_set_rate(pcmHandle, hwparams, pRate, 0);
        if (r < 0)
                {
                PERROR("Invalid sample rate (%d Hz) to bus %s . ALSA ERROR : %s", pRate, cBusFullReference, snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        // the period/buffer crap setup process
        r = snd_pcm_hw_params_set_periods(pcmHandle, hwparams, numPeriods, 0) ;
        if (r < 0)
                {
                PERROR("Cannot set periods count to %d on bus %s. ALSA ERROR : %s", numPeriods, cBusFullReference, snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        r = snd_pcm_hw_params_get_period_size_min (hwparams, &minPeriodSize, &dir );
        if (r<0)
                {
                PERROR("Cannot get minimum period size on bus %s . ALSA ERROR : %s", cBusFullReference, snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }
        r = snd_pcm_hw_params_get_period_size_max (hwparams, &maxPeriodSize, &dir );
        if (r<0)
                {
                PERROR("Cannot get maximum period size on bus %s . ALSA ERROR : %s", cBusFullReference, snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        PMESG2("Deteceted min/max period size : %d - %d",minPeriodSize,maxPeriodSize);

        periodSize = fragmentSize / blockSize;
        // TODO : check if periodSize is allow (is within mix/max)

        triedPeriodSize = (int) periodSize;

        if (stream==MustuxAudioDeviceMapper::PLAYBACK)
                {
                PMESG2("Trying OPEN PLAYBACK BUS : fps=%d  fragmentSize=%d  periods=%d  periodSize=%d",MustuxAudioDeviceMapper::get_fps(), fragmentSize, numPeriods, periodSize);                int rp2 = snd_pcm_hw_params_set_period_size (pcmHandle, hwparams, periodSize, 0);
                r = snd_pcm_hw_params_set_period_size (pcmHandle, hwparams, periodSize, 0);
                if (r < 0)
                        {
                        PERROR("setting period size to %d on bus %s. ALSA ERROR : %s", periodSize,cBusFullReference,snd_strerror(r));
                        goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                        }
                }
        else
                {
                PMESG2("Trying OPEN MustuxAudioDeviceMapper::CAPTURE BUS : fps=%d  fragmentSize=%d  periods=%d  periodSize=%d",MustuxAudioDeviceMapper::get_fps(), fragmentSize, numPeriods, periodSize);
                r = snd_pcm_hw_params_set_period_size_near (pcmHandle, hwparams, &periodSize, &dir);
                if (r < 0)
                        {
                        PERROR("setting period size to %d on bus %s. ALSA ERROR : %s", periodSize,cBusFullReference, snd_strerror(r));
                        goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                        }
                }

        r = snd_pcm_hw_params_get_period_size (hwparams, &periodSize, &dir);
        if (r < 0)
                {
                PERROR("Could not confirm actual periodSize value after setting bus %s . ALSA ERROR : ",cBusFullReference, snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        if (triedPeriodSize!=periodSize)
                {
                PMESG2("Could not set periodSize to %d . ALSA let me use only : %d . Continuing anyway...",triedPeriodSize, periodSize);
                }


        // write the parameters to descriptor
        r = snd_pcm_hw_params(pcmHandle, hwparams);
        if (r < 0)
                {
                PERROR("Unable to set hw params : %s",snd_strerror(r));
                goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
                }

        transferSize = periodSize * blockSize;
        transferBuffer=new char[transferSize];
        bzero(transferBuffer, transferSize);
#ifdef THREADED_TRANSFER
        safeTransferBuffer=new char[transferSize];
        bzero(safeTransferBuffer, transferSize);
#endif
        if (stream==MustuxAudioDeviceMapper::PLAYBACK)
                {
                MustuxAudioDeviceMapper::set_common_transfer_size(transferSize); // ATTENTION : will be overwritten everytime a playback bus is open. But no problem, since it is supposed to overwrite always with same value
                if (channels==1)
                        status=PLAYING_TO_MONO;
                else
                        status=PLAYING_TO_STEREO; // shouldnt it have status for "playing to left only/ playing to right only" situatiosn ???
#ifdef THREADED_TRANSFER
                transferThread->start();
#endif
                }
        else
                {
                if (whichChannels==MustuxAudioDeviceMapper::ONLY_LEFT_CHANNEL)
                        status=CAPTURING_FROM_LEFT_CHANNEL;
                else if (whichChannels==MustuxAudioDeviceMapper::ONLY_RIGHT_CHANNEL)
                        status=CAPTURING_FROM_RIGHT_CHANNEL;
                else if (whichChannels==MustuxAudioDeviceMapper::STEREO)
                        status=CAPTURING_FROM_STEREO;
                else if (whichChannels==MustuxAudioDeviceMapper::MONO)
                        status=CAPTURING_FROM_MONO;
                }

        if (status!=CLOSED)
                {
                PMESG2("*** %s BUS %s OPEN : %d Hz, %s, %d bits periodSize=%d ***",stream==MustuxAudioDeviceMapper::CAPTURE?"MustuxAudioDeviceMapper::CAPTURE":"PLAYBACK",cBusFullReference,pRate ,channels==1?"MONO":"STEREO",pBitDepht,(int) periodSize);
                }
        else
                {
                PERROR("Status is still closed, at this point ??!?!?!");
                }

        PEXIT2;
        return 1;

ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION:
        if (snd_pcm_close(pcmHandle)<0)
                {
                PERROR ("Could not close pcmHandle after previous ALSA error");
                }
        // SHOULDNT I CALL snd_pcm_hw_params_free(hwparams) now ?
        // If I try I get a segfault.. but why.. ?
        PEXIT2;
        return r;
        }



int Bus::transfer(int pBytesToTransfer)
        {
        PENTER3;
        int r;
        int i=0;
        bytesToTransfer=pBytesToTransfer;
        bytesTransferred=0;
        if (stream==MustuxAudioDeviceMapper::CAPTURE)
                {
                // bytesToTransfer parameters is ignored on capture buses
                r = snd_pcm_readi(pcmHandle, transferBuffer, periodSize);
                if (r < 0) // buffer underrun
                        {
                        PWARN2("Underrun on bus %s",(const char*) descriptor.latin1());
                        snd_pcm_prepare(pcmHandle);
                        transferError = MustuxAudioDeviceMapper::ERROR_UNDERRUN;
                        PEXIT3;
                        return transferError;
                        }
                /* what the heck is this ?
                if (r == -EAGAIN)
                */
                transferError = 0;
                bytesTransferred = r * blockSize;
                }
        else

#ifdef THREADED_TRANSFER
                {
                //printf("*** %s Waiting current transfer ...\n",descriptor.ascii());
                while (transferThread->transferring) // wait until current transf (if any( finishes
                        usleep(500);
                //printf("*** %s Current transfer finished! \n",descriptor.ascii());
                memcpy(safeTransferBuffer,transferBuffer,transferSize);
                transferThread->transferring=true;
                //printf("*** %s New transfer order liberated...\n",descriptor.ascii());
                }
#else
                {
                int periodsTransferred;
                if (bytesToTransfer==-1)
                        periodsTransferred = snd_pcm_writei(pcmHandle, transferBuffer, periodSize);
                else
                        {
                        if (bytesToTransfer > periodSize)
                                {
                                PWARN2("Trying to transfer more than bus buffer size!");
                                bytesToTransfer = periodSize;
                                }
                        partialPeriodSize = bytesToTransfer / blockSize;
                        periodsTransferred = snd_pcm_writei(pcmHandle, transferBuffer, partialPeriodSize);
                        }
                if (periodsTransferred < 0) // buffer underrun
                        {
                        PWARN2("Underrun on bus %s",(const char*) descriptor.latin1());
                        snd_pcm_prepare(pcmHandle);
                        transferError = MustuxAudioDeviceMapper::ERROR_UNDERRUN;
                        PEXIT3;
                        return transferError;
                        }
                /* what the heck is this ?
                if (r == -EAGAIN)
                        return 1;
                */
                transferError=0;
                bytesTransferred = periodsTransferred * blockSize;
                }
#endif

        PEXIT3;
        return bytesTransferred;
        }




void Bus::clean_buffer()
        {
        PENTER3;
        bzero(transferBuffer, transferSize);
#ifdef THREADED_TRANSFER
        bzero(safeTransferBuffer, transferSize);
#endif
        PEXIT3;
        }



int Bus::close()
        {
        PENTER2;
#ifdef THREADED_TRANSFER
        transferThread->transferring=false;
        transferThread->transferThreadRunning=false;
        transferThread->wait();
#endif
        if (status==CLOSED)
                {
                PMESG2("Bus %s Seems to be already closed. Ignoring close call...",descriptor.latin1());
                PEXIT2;
                return 1;
                }
        if (transferBuffer)
                delete transferBuffer;
#ifdef THREADED_TRANSFER
        if (safeTransferBuffer)
                delete safeTransferBuffer;
        safeTransferBuffer=(char*) 0;
#endif
        transferBuffer=(char*) 0;

        if ( snd_pcm_close(pcmHandle) <0 )
                {
                PERROR("Cannot close Bus %s", (const char*) fullName );
                PEXIT2;
                return -1;
                }
        status=CLOSED;
        PMESG2(" BUS %s CLOSED", (const char*) fullName );
        // SHOULDNT I CALL THIS ? (it crashes !) snd_pcm_hw_params_free(hwparams);
        PEXIT2;
        return 1;
        }


int Bus::get_transfer_size()
        {
        return transferSize;
        }


char* Bus::get_transfer_buffer(int whichChannels)
        {
        return transferBuffer; // IMPROVE ME.. Mustux consider whichChannels some way...
        }


bool Bus::is_stereo()
        {
        return stereoBus;
        }


QString Bus::get_short_name()
        {
        return shortName;
        }

QString Bus::get_descriptor()
        {
        return descriptor;
        }

QString Bus::get_full_name()
        {
        return fullName;
        }

bool Bus::allocated_for_playback()
        {
        bool b = ( (status==PLAYING_TO_STEREO) || (status==PLAYING_TO_LEFT_CHANNEL) || (status==PLAYING_TO_RIGHT_CHANNEL));
        return b;
        }

bool Bus::allocated_for_capture()
        {
        return ( (status==CAPTURING_FROM_RIGHT_CHANNEL) || (status==CAPTURING_FROM_LEFT_CHANNEL) || (status==CAPTURING_FROM_STEREO));
        }

QString Bus::scbModeName[Bus::MAX_SCB_MODES] =
                               {"22050 MONO 16 bits",
                                "22050 STEREO 16 bits",
                                "32000 MONO 16 bits",
                                "32000 STEREO 16 bits",
                                "44100 MONO 16 bits",
                                "44100 STEREO 16 bits",
                                "44100 MONO 24 bits",
                                "44100 STEREO 24 bits",
                                "48000 MONO 16 bits",
                                "48000 STEREO 16 bits",
                                "48000 MONO 24 bits",
                                "48000 STEREO 24 bits",
                                "96000 MONO 16 bits",
                                "96000 STEREO 16 bits",
                                "96000 MONO 24 bits",
                                "96000 STEREO 24 bits"};

int Bus::scbModeRate[Bus::MAX_SCB_MODES] = { 22050, 22050, 32000, 32000, 44100, 44100, 44100, 44100, 48000, 48000, 48000, 48000, 96000, 96000, 96000, 96000};
int Bus::scbModeChannels[Bus::MAX_SCB_MODES] = { 1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2 };
int Bus::scbModeBitDepht[Bus::MAX_SCB_MODES] = { 16,16,16,16,16,16,24,24,16,16,24,24,16,16,24,24};


// eof

