/**************************************************************************************

        MUSTUXLIB - THE COMMON LIBRARY FOR ALL MUSTUX APPLICATIONS
        AUTHOR : See AUTHORS file for details

        This software is distributed under the terms of the GNU General Public License
        as specified in the COPYING file.

***************************************************************************************/

#ifndef MUSTUXAUDIOBUS_HH
#define MUSTUXAUDIOBUS_HH

#define ALSA_PCM_NEW_HW_PARAMS_API
#include <alsa/asoundlib.h>

#include <qstring.h>
#include <qthread.h>

// These are auxiliary classes, not meant to be used directly by user.

//#define THREADED_TRANSFER
#ifdef THREADED_TRANSFER
class Bus;

class TransferThread : public QThread
        {
        public:
                TransferThread (Bus* pBus);
                void run();
                Bus* bus;
                bool transferThreadRunning;
                bool transferring;
        };
#endif

class Bus
        {
        public:
                static const int MAX_SCB_MODES = 16;
                static const int CLOSED = 0;
                static const int PLAYING_TO_MONO = 1;
                static const int PLAYING_TO_STEREO = 2;
                static const int PLAYING_TO_LEFT_CHANNEL  = 3;
                static const int PLAYING_TO_RIGHT_CHANNEL  = 4;
                static const int CAPTURING_FROM_MONO = 5;
                static const int CAPTURING_FROM_STEREO = 6;
                static const int CAPTURING_FROM_LEFT_CHANNEL = 7;
                static const int CAPTURING_FROM_RIGHT_CHANNEL = 8;
                static QString scbModeName[MAX_SCB_MODES];
                static int scbModeRate[MAX_SCB_MODES];
                static int scbModeChannels[MAX_SCB_MODES];
                static int scbModeBitDepht[MAX_SCB_MODES];


                Bus(int pParentCardId, int pId, int pStream, QString desc);
                ~Bus();
                void probe_valid_modes();
                int supports(int pRate, int pBitDepht, int pChannels, int pStream);
                bool valid_mode(int pRate, int pBitDepht, int pChannels);
                int open(int pRate, int pBitDepht, int whichChannels,bool block=true);
                int transfer(int bytesToTransfer=-1);
                void clean_buffer();
                int close();
                char* get_transfer_buffer(int whichChannels);
                int get_transfer_size();
                bool allocated_for_playback();
                bool allocated_for_capture();
                bool is_stereo();
                QString get_short_name();
                QString get_full_name();
                QString get_descriptor();

#ifdef THREADED_TRANSFER
                TransferThread* transferThread;
                char* safeTransferBuffer;
#endif

                int transferError;
                int bytesToTransfer;
                QString shortName;
                QString fullName;
                int stream;   // CAPTURE or PLAYBACK
                int parentCardId;
                int id;

                int status;
                bool stereoBus;
                int blockSize;

                snd_pcm_t*              pcmHandle;
                snd_pcm_hw_params_t*    hwparams;
                snd_pcm_format_t format;

                QString descriptor; // those "hw:x,x" ALSA strings
                bool isValidMode[MAX_SCB_MODES];

                char* transferBuffer;
                int transferSize;
                int partialPeriodSize;
                int bytesTransferred;
                snd_pcm_uframes_t periodSize;

        };


#endif

